"""Train with gradient caching. Useful for training a large model that requires lots of GPU memory.

Initially I tried to make a wrapper for Model without modifing the Trainer class. However, it required a backward loop in a training_step(), which is bit counterintuitive.
Also having a backward pass in training_step() will make a conflict with plugins that have callbacks for backward pass.
For example, the AMP plugin might require special implementation for gradient caching.

"""
import contextlib
import copy
import dataclasses
import logging
import irisml.core
import torch
from .train import Task as TrainTask
from .train.ddp_utils import all_gather, all_gather_object

logger = logging.getLogger(__name__)


class Task(TrainTask):
    """Train a model using gradient cache. Useful for contrastive learning with a large model.

    This trainer expects the model to return a single FloatTensor or a tuple of FloatTensors or a scalar Tensor.
    If a scalar Tensor is returned, the trainer assumes it is a fixed value in an iteration regardless of inputs.

    Config:
        sub_batch_size (int): Sub batch size for each forward/backward pass. batch_size must be divisible by sub_batch_size.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Config(TrainTask.Config):
        sub_batch_size: int = 1

    def execute(self, inputs):
        if self.config.batch_size % self.config.sub_batch_size != 0:
            raise ValueError(f"batch_size must be divisible by sub_batch_size. batch_size={self.config.batch_size}, sub_batch_size={self.config.sub_batch_size}")
        return super().execute(inputs)

    def get_trainer(self, parameters):
        parameters['sub_batch_size'] = self.config.sub_batch_size
        if self.config.num_processes == 1:
            return GradientCachingTrainer(**parameters)
        else:
            return GradientCachingDDPTrainer(num_processes=self.config.num_processes, **parameters)


class GradientCachingMixin:
    class ModelWrapper:
        def __init__(self, model, device):
            self._model = model
            self._device = device
            self._criterion = model.criterion
            self._has_multiple_features = None

        def __call__(self, inputs, rng_states=None):
            """
            Args:
                inputs (Any): inputs to the model
                rng_states (List): Random Number Generator state to use. Default: None

            Returns:
                (features, rng_states)
            """
            current_rng_states = rng_states or self._get_rng_states()
            with self._use_rng_states(rng_states):
                features = self._model(inputs)

            if self._has_multiple_features is None:
                self._has_multiple_features = isinstance(features, (list, tuple))
            if not self._has_multiple_features:
                features = [features]

            return features, current_rng_states

        def calculate_loss(self, features, targets):
            if not self._has_multiple_features:
                features = features[0]
            return self._criterion(features, targets)

        def _get_rng_states(self):
            states = [torch.get_rng_state()]
            if self._device.type == 'cuda':
                states.append(torch.cuda.get_rng_state(self._device))
            return states

        @contextlib.contextmanager
        def _use_rng_states(self, rng_states):
            if rng_states:
                devices = [self._device] if self._device.type == 'cuda' else None
                with torch.random.fork_rng(devices):
                    torch.set_rng_state(rng_states[0])
                    if len(rng_states) > 1:
                        torch.cuda.set_rng_state(rng_states[1], self._device)
                    yield
            else:
                yield

    def __init__(self, sub_batch_size, **kwargs):
        super().__init__(**kwargs)
        self._sub_batch_size = sub_batch_size
        assert self.criterion, "The model must have a criterion property to use gradient caching."

    def train_epoch(self, dataloader, epoch):
        if not self._plugins.on_train_epoch_start(self, self.model, epoch):
            return False  # Early stopping

        model = GradientCachingMixin.ModelWrapper(self.model, self.device)

        for batch_index, batch in enumerate(dataloader):
            self._optimizer.zero_grad()
            batch = self._plugins.on_train_batch_start(self, self.model, batch, batch_index)
            batch = self._to_device(batch)
            sub_batches = list(zip(self._split_tensor_or_list(batch[0]), self._split_tensor_or_list(batch[1])))

            # Run forward pass for the sub batches.
            with self._plugins.forward_context():
                with torch.no_grad():
                    results = [model(b[0]) for b in sub_batches]
                    features_local = [r[0] for r in results]
                    rng_states = [r[1] for r in results]
                    # Note that a feature vector can contain a scalar tensor. In such case, uses the first scalar value since the trainer assumes all sub_batch returns a same value.
                    features = [torch.cat(f, dim=0) if f[0].shape else f[0] for f in zip(*features_local)]

            # Set requires_grad in order to calculate gradients for features on this device.
            for f in features:
                f.requires_grad_()

            features_all = [self._all_gather(f) for f in features]
            targets_all = self._all_gather(batch[1])
            loss = model.calculate_loss(features_all, targets_all)

            # Run backward pass to calculate gradients for 'features'.
            loss = self._plugins.on_train_backward_start(self, self.model, loss)
            loss.backward()
            self._plugins.on_train_backward_end(self, self.model)

            # Copy the gradients of the feature vectors. Deep-copied so that the subsequent backward pass doesn't affect the cache
            gradient_cache = [self._split_tensor_or_list(copy.deepcopy(f.grad.detach())) for f in features]

            for i, sub_batch in enumerate(sub_batches):
                # Run forward pass again, with gradients this time.
                with self._plugins.forward_context():
                    features_local_recalculated = model(sub_batch[0], rng_states[i])[0]
                    assert len(features_local_recalculated) == len(gradient_cache)

                maybe_no_sync = self.model.no_sync() if hasattr(model, 'no_sync') and i + 1 == len(sub_batches) else contextlib.nullcontext
                with maybe_no_sync():
                    # Run backward pass using the cached gradients for each features.
                    for j, f in enumerate(features_local_recalculated):
                        if not torch.allclose(f, features_local[i][j]):
                            diff = torch.max(torch.abs(f - features_local[i][j]))
                            logger.warning(f"The recalculated feature({i=}, {j=}) has different values. {diff=}")
                        gradient = gradient_cache[j][i] if f.shape else (gradient_cache[j] / len(sub_batches))
                        f.backward(gradient=gradient)

            optimizer = self._plugins.on_optimizer_step_start(self, self.model, self._optimizer)
            optimizer.step()
            self._plugins.on_train_batch_end(self, self.model, loss, batch, batch_index)
            self._lr_scheduler.step()

        self._plugins.on_train_epoch_end(self, self.model, epoch)

    def _all_gather(self, tensor_or_list):
        """Gather tensors from all devices. Keep the local tensors as they are."""
        if not torch.distributed.is_initialized():
            return tensor_or_list

        if torch.is_tensor(tensor_or_list):
            if not tensor_or_list.shape:
                return tensor_or_list  # TODO: We have an assumption that a sclar tensor is same on all devices.
            return all_gather(tensor_or_list)
        else:
            assert isinstance(tensor_or_list, list)
            return all_gather_object(tensor_or_list)

    def _split_tensor_or_list(self, tensor_or_list):
        if torch.is_tensor(tensor_or_list):
            if not tensor_or_list.shape:
                return tensor_or_list  # Returns a scalar value as-is.
            assert len(tensor_or_list) % self._sub_batch_size == 0, f"len(tensor_or_list) must be divisible by sub_batch_size. {len(tensor_or_list)}, {self._sub_batch_size}"
            return torch.split(tensor_or_list, self._sub_batch_size)
        elif torch.is_tensor(tensor_or_list[0]):
            assert all(len(t) % self._sub_batch_size == 0 for t in tensor_or_list)
            return list(zip(*[torch.split(t, self._sub_batch_size) for t in tensor_or_list]))
        else:
            assert isinstance(tensor_or_list, list)
            assert len(tensor_or_list) % self._sub_batch_size == 0, f"len(tensor_or_list) must be divisible by sub_batch_size. {len(tensor_or_list)}, {self._sub_batch_size}"
            num_chunks = self.config.batch_size // self._sub_batch_size
            bs = self._sub_batch_size
            return [tensor_or_list[i * bs: (i + 1) * bs] for i in range(num_chunks)]


class GradientCachingTrainer(GradientCachingMixin, irisml.tasks.train.Trainer):
    pass


class GradientCachingDDPTrainer(GradientCachingMixin, irisml.tasks.train.DDPTrainer):
    pass



import copy
import dataclasses
import torch.nn
import irisml.core


class Task(irisml.core.TaskBase):
    """Split a image-text model into an image model and a text model.

    Inputs:
        model (torch.nn.Module): An input model. It must have 'image_model' and 'text_model' attributes.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Outputs:
        image_model: torch.nn.Module = None
        text_model: torch.nn.Module = None

    def execute(self, inputs):
        image_model = torch.nn.Sequential(copy.deepcopy(inputs.model.image_model), copy.deepcopy(inputs.model.image_projection))
        text_model = torch.nn.Sequential(copy.deepcopy(inputs.model.text_model), copy.deepcopy(inputs.model.text_projection))
        return self.Outputs(image_model, text_model)

    def dry_run(self, inputs):
        return self.execute(inputs)



"""Mixup (https://arxiv.org/abs/1710.09412)

For classification_multiclass, targets will be updated to one-hot vectors, then mixup will be applied.

For object_detection, bounding boxes are simply concatenated without changing box probabilities.
"""

import logging
import typing
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Delegate:
    def __init__(self, alpha, num_classes):
        self._beta_distribution = torch.distributions.beta.Beta(alpha, alpha)
        self.num_classes = num_classes

    def get_lambda(self):
        return self._beta_distribution.sample()

    def __call__(self, batch):
        inputs, targets = batch
        indices = torch.randperm(len(inputs), device=inputs.device, dtype=torch.long)
        coefficient = self.get_lambda()
        return self.mixup(batch, indices, coefficient)

    def mixup(self, batch, indices, coefficient):
        raise NotImplementedError


class ClassificationMulticlassDelegate(Delegate):
    def mixup(self, batch, indices, coefficient):
        inputs, targets = batch
        inputs = inputs * coefficient + inputs[indices] * (1 - coefficient)
        onehot_targets = torch.nn.functional.one_hot(targets, self.num_classes)
        print("targets")
        print(targets)
        targets = onehot_targets * coefficient + onehot_targets[indices] * (1 - coefficient)
        print(onehot_targets)
        return inputs, targets


class ClassificationMultilabelDelegate(Delegate):
    def mixup(self, batch, indices, coefficient):
        inputs, targets = batch
        inputs = inputs * coefficient + inputs[indices] * (1 - coefficient)
        targets = targets * coefficient + targets[indices] * (1 - coefficient)
        return inputs, targets


class ObjectDetectionDelegate(Delegate):
    def mixup(self, batch, indices, coefficient):
        inputs, targets = batch
        inputs = inputs * coefficient + inputs[indices] * (1 - coefficient)
        targets = [torch.cat((t, targets[indices[i]])) for i, t in enumerate(targets)]
        return inputs, targets


class Plugin(PluginBase):
    def __init__(self, task_type: typing.Literal['classification_multiclass', 'classification_multilabel', 'object_detection'], alpha: float, num_classes: int):
        if task_type == 'classification_multiclass':
            self._delegate = ClassificationMulticlassDelegate(alpha, num_classes)
        elif task_type == 'classification_multilabel':
            self._delegate = ClassificationMultilabelDelegate(alpha, num_classes)
        elif task_type == 'object_detection':
            self._delegate = ObjectDetectionDelegate(alpha, num_classes)
        else:
            raise ValueError(f"Unsupported task type: {task_type}")

        logger.info(f"Mixup is enabled with {alpha=}, {task_type=}, {num_classes=}")

    def on_train_batch_start(self, trainer, model, batch, batch_index):
        if len(batch[0]) == 1:
            return batch  # Do nothing if batch_size is 1.
        return self._delegate(batch)



import dataclasses
import logging
import torch
import PIL.Image
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Convert an object detection dataset into a classification dataset."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        return self.Outputs(ClassificationFromDetectionDataset(inputs.dataset))

    def dry_run(self, inputs):
        return self.execute(inputs)


class ClassificationFromDetectionDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self._dataset = dataset
        num_boxes = [len(targets) for _, targets in dataset]
        self._index_mappings = []
        for i, n in enumerate(num_boxes):
            self._index_mappings.extend([(i, j) for j in range(n)])

        logger.info(f"Created a classificaiton dataset with {len(self._index_mappings)} samples. The source dataset has {len(self._dataset)} samples.")
        assert len(self._index_mappings) == sum(num_boxes)

    def __len__(self):
        return len(self._index_mappings)

    def __getitem__(self, index):
        image_index, box_index = self._index_mappings[index]
        image, targets = self._dataset[image_index]
        assert isinstance(image, PIL.Image.Image)
        w, h = image.size
        box = targets[box_index]
        cropped = image.crop((int(w * box[1]), int(h * box[2]), int(w * box[3]), int(h * box[4])))
        return cropped, box[0].int()



import logging
import sys
import torch.distributed
import tqdm
from .plugin_base import PluginBase


class TqdmLoggingHandler(logging.StreamHandler):
    def emit(self, record):
        msg = self.format(record)
        tqdm.tqdm.write(msg, file=self.stream)
        self.flush()


def _is_console_handler(handler):
    return isinstance(handler, logging.StreamHandler) and handler.stream in (sys.stdout, sys.stderr)


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        self._disabled = torch.distributed.is_initialized() and torch.distributed.get_rank() != 0
        if not self._disabled:
            self._training_progressbar = tqdm.tqdm(total=trainer.num_epochs, position=0, disable=None)
            # Replace the logging stream handlers
            self._original_handlers = logging.root.handlers
            stream_handlers = [h for h in logging.root.handlers if _is_console_handler(h)]
            if stream_handlers:
                tqdm_handler = TqdmLoggingHandler()
                tqdm_handler.setFormatter(stream_handlers[0].formatter)
                tqdm_handler.stream = stream_handlers[0].stream
                logging.root.handlers = [h for h in logging.root.handlers if not _is_console_handler(h)] + [tqdm_handler]

    def on_train_end(self, trainer, model):
        if not self._disabled:
            self._training_progressbar.close()
            logging.root.handlers = self._original_handlers

    def on_train_epoch_end(self, trainer, model, epoch_index):
        if not self._disabled:
            self._epoch_progressbar.close()
            self._training_progressbar.update(1)

    def on_train_epoch_start(self, trainer, model, epoch_index):
        if not self._disabled:
            self._epoch_progressbar = tqdm.tqdm(total=len(trainer.train_dataloader), position=1, leave=None, disable=None)
        return True

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        if not self._disabled:
            self._epoch_progressbar.update(1)



import datetime
import torch.utils.tensorboard
from .plugin_base import PluginBase


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        current_time = datetime.datetime.now().strftime('%m%d%H%M%S')
        self._writer = torch.utils.tensorboard.SummaryWriter(f'logs/{current_time}')
        self._counter = 0

    def on_train_end(self, trainer, model):
        self._writer.close()

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        for i, lr in enumerate(g['lr'] for g in trainer.optimizer.param_groups):
            self._add_scalar(f'Lr{i}', lr, False)
        self._add_scalar('Loss/train', float(loss.item()))

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        self._add_scalar('Loss/validation', float(loss.item()))

    def _add_scalar(self, name, value, increment=True):
        self._writer.add_scalar(name, value, self._counter)
        if increment:
            self._counter += 1



import contextlib


class PluginList:
    """Provides a single Plugin interface for a list of plugins."""
    def __init__(self, plugins):
        self._plugins = plugins

    def on_train_start(self, trainer, model):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_start'):
                plugin.on_train_start(trainer, model)

    def on_train_end(self, trainer, model):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_end'):
                plugin.on_train_end(trainer, model)

    def on_train_epoch_start(self, trainer, model, epoch_index):
        flag = True
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_epoch_start'):
                flag &= plugin.on_train_epoch_start(trainer, model, epoch_index)
        return flag

    def on_train_epoch_end(self, trainer, model, epoch_index):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_epoch_end'):
                plugin.on_train_epoch_end(trainer, model, epoch_index)

    def on_train_batch_start(self, trainer, model, batch, batch_index):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_batch_start'):
                batch = plugin.on_train_batch_start(trainer, model, batch, batch_index)
        return batch

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_batch_end'):
                plugin.on_train_batch_end(trainer, model, loss, batch, batch_index)

    def on_train_backward_start(self, trainer, model, loss):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_backward_start'):
                loss = plugin.on_train_backward_start(trainer, model, loss)
        return loss

    def on_train_backward_end(self, trainer, model):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_train_backward_end'):
                plugin.on_train_backward_end(trainer, model)

    def on_optimizer_step_start(self, trainer, model, optimizer):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_optimizer_step_start'):
                optimizer = plugin.on_optimizer_step_start(trainer, model, optimizer)
        return optimizer

    def on_validation_start(self, trainer, model):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_validation_start'):
                plugin.on_validation_start(trainer, model)

    def on_validation_end(self, trainer, model):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_validation_end'):
                plugin.on_validation_end(trainer, model)

    def on_validation_batch_start(self, trainer, model, batch, batch_index):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_validation_batch_start'):
                batch = plugin.on_validation_batch_start(trainer, model, batch, batch_index)
        return batch

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        for plugin in self._plugins:
            if hasattr(plugin, 'on_validation_batch_end'):
                plugin.on_validation_batch_end(trainer, model, loss, batch, batch_index)

    @contextlib.contextmanager
    def forward_context(self):
        with contextlib.ExitStack() as stack:
            for plugin in self._plugins:
                if hasattr(plugin, 'forward_context'):
                    ctx = plugin.forward_context()
                    if ctx:
                        stack.enter_context(ctx)
            yield
